/* -*- mode: c++; c-basic-offset: 3; -*- */
// WSTREAM Top level function for the Omega Pipeline burst search
//
// WSTREAM applies the discrete Q transform to search for statistically
// significant transient events in data from interferometric
// gravitational-wave detectors.
//
// usage: wstream(startTime, stopTime, parameterFile, frameCacheFile, 
//                outputDirectory, debugLevel)
//
//   startTime           gps start time of analysis
//   stopTime            gps stop time of analysis
//   parameterFile       parameter file
//   frameCacheFile      readframedata formatted frame cache file
//   outputDirectory     directory to write results
//   debugLevel          verboseness level of debug output
//
// If the specified stop time is less than the specified start time, it is
// instead assumed to be the duration of the analysis.  Non-integer start
// and stop times are truncated to the nearest integer.
//
// If no output directory is specified, WSTREAM places the resulting trigger
// and event files in a subdirectory of the current directory, named after 
// the specified segment:
//
//   segments/
//     <startTime>-<stopTime>/
//       livetime.txt
//       <channelName1>.txt
//       <channelName2>.txt
//       
//
// If an outputDirectory is specified, then all output is written to that
// directory instead.
//
// If no parameter file or frame cache file is specified, WSTREAM looks
// for the files parameters.txt or framecache.txt in the current directory.
//
// The specified debugLevel controls the amount of detail in the output log.
// A debugLevel of unity is assumed by default.
//
// See also WBLOCK, WEVENT

// Authors:
// Shourov K. Chatterji <shourov@ligo.caltech.edu>
// Leo C. Stein <lstein@ligo.mit.edu>
// Jameson Graef Rollins <jrollins@phys.columbia.edu>

#include "wtypes.hh"
#include "wtile.hh"
#include "wchunk.hh"
#include "woutput.hh"
#include "wparameters.hh"
#include "matlab_fcs.hh"

#include "EggTimer.hh"
#include "MultiDacc.hh"
#include "SigFlag.hh"

#include <iostream>
#include <iomanip>
#include <sstream>
#include <cstdlib>

using namespace wpipe;
using namespace std;

//==================================  Search function prototype
void
wstream(const std::string& params, const std::string& cache,
	const std::string& outdir, int debugLevel);


//==================================  Main function
void
syntax(void) {
   cerr << "Command line syntax: " << endl;
   cerr << "  wstream <par-file> <frame-cache> <temp-dir> <debug-level>"
	<< endl;
}


//==================================  Main function
//
//  Command line:
//    wstream <par-file> <frame-cache> <temp-dir> <debug-level>
//
int
main(int argc, const char* argv[]) {
   if (argc < 3) {
      cerr << "insufficient parameters:" << endl;
      syntax();
      return 1;
   }
   string parFile = argv[1];
   string frameCache = argv[2];
   string outDir;
   if (argc > 3) outDir = argv[3];
   int debugLevel = 1;
   if (argc > 4) debugLevel = strtol(argv[4], 0, 0);

   try {
      wstream(parFile, frameCache, outDir, debugLevel);
   } catch (std::exception& e) {
      cerr << "Caught exception: " << e.what() << endl;
   }
}

//==================================  Event search function.
void
wstream(const std::string& params, const std::string& cache,
	const std::string& outdir, int debugLevel) {
   SigFlag term(SIGTERM);
   term.add(SIGINT);

   /////////////////////////////////////////////////////////////////////////////
   //                           start analysis timer                          //
   /////////////////////////////////////////////////////////////////////////////

   // start time of analysis
   EggTimer analysisTimer;

   /////////////////////////////////////////////////////////////////////////////
   //                         parse command line arguments                    //
   /////////////////////////////////////////////////////////////////////////////
   // apply default arguments
   string parameterFile = params;
   if (params.empty()) {
      parameterFile = "parameters.txt";
   }

   string frameCacheFile = cache;
   if (frameCacheFile.empty()) {
      frameCacheFile = "framecache.txt";
   }

   string outputDirectory = outdir;

   // create output path if (not specified
   ostringstream oss;
   if (outputDirectory.empty()) {
      oss << "./segments/" << Now().getS() << "-online";
      outputDirectory = oss.str();
      oss.str("");
      // path to segment output directory
   }

   // check if specified path is absolute
   else if (outputDirectory[0] != '/') {
      // make explicitly relative to current directory if (not absolute
      outputDirectory = string("./") + outputDirectory;
   }

   ////////////////////////////////////////////////////////////////////////////
   //                              write header                              //
   ////////////////////////////////////////////////////////////////////////////

   // report analysis name, time and command line arguments
   if (debugLevel >= 1) {
      cout << ">> DMT Omega streaming analysis" << endl;
      cout << "Run by " << getenv("USER") << " on " << datestr(0, 29)
	   << " at " << datestr(0, 13) << endl;
      //cout << "  analysis start time:     " << startTimeString << endl;
      //cout << "  analysis stop time:      " << stopTimeString << endl;
      cout << "  parameter file:          " << parameterFile << endl;
      cout << "  framecache file:         " << frameCacheFile << endl;
      cout << "  output directory:        " << outputDirectory << endl;
      cout << "  debug level:             " << debugLevel << endl;
   }

   /////////////////////////////////////////////////////////////////////////////
   //                               read parameters                           //
   /////////////////////////////////////////////////////////////////////////////

   // report status
   wlog(debugLevel, 1, "reading parameter file");

   // read parameters
   wparameters pars(parameterFile, debugLevel);
   pars.display(cout);

   /////////////////////////////////////////////////////////////////////////////
   //                              tile signal space                          //
   /////////////////////////////////////////////////////////////////////////////

   // report status
   wlog(debugLevel, 1, "tiling search space");

   // generate Q transform tiling
   double fSample = pars.sampleFrequency;
   wtile tiling(pars.blockDuration, pars.qRange, pars.frequencyRange, 
		fSample, pars.maximumMismatch, 
		pars.highPassCutoff, pars.lowPassCutoff, 
		pars.whiteningDuration, pars.transientFactor);

   if (debugLevel >=2) {
      cout << "  whiteningDuration:   " << tiling.whiteningDuration() << endl;
      cout << "  transientDuration:   " << tiling.transientDuration() << endl;
   }

   /////////////////////////////////////////////////////////////////////////////
   //                              partition segment                          //
   /////////////////////////////////////////////////////////////////////////////

   // report status
   wlog(debugLevel, 1, "portioning segment");

   //----------------------------------  get max, min time shift (sample
   //                                    alignment forced in pars)
   double maximumTimeShift(0), minimumTimeShift(0);
   if (!pars.timeShifts.empty()) {
      maximumTimeShift = pars.timeShifts[0];
      minimumTimeShift = pars.timeShifts[0];
      size_t N = pars.timeShifts.size();
      for (size_t i=1; i<N; i++) {
	 if (pars.timeShifts[i] > maximumTimeShift) {
	    maximumTimeShift = pars.timeShifts[i];
	 } else if (pars.timeShifts[i] < minimumTimeShift) {
	    minimumTimeShift = pars.timeShifts[i];
	 }
      }
   }

   // determine minimum block overlap
   double minimumBlockOverlap = 2 * tiling.transientDuration() + 
                                pars.extraBlockOverlap;

   if (debugLevel >=2) {
      cout << "  minimumBlockOverlap: " << minimumBlockOverlap << endl;
   }

   // total segment livetime loss due to time shifts
   double timeShiftLoss = maximumTimeShift - minimumTimeShift;

   // if (segment duration is less than block duration
   //if (double(segmentDuration) - timeShiftLoss < pars.blockDuration) {
   //   // write error to log file
   //   error("segment too short to analyze");
   //}

   double blockOverlap = minimumBlockOverlap;
   if (blockOverlap > floor(blockOverlap)) blockOverlap += 1.0;

   // name of livetime file
   string livetimeFile = outputDirectory + "/livetime.txt";

   // report segment information
   if (debugLevel >= 1) {
      cout << "  block duration:      " << pars.blockDuration << " seconds" 
	   << endl;
      cout << "  time shift loss:     " << timeShiftLoss   << " seconds"<< endl;
      cout << "  block overlap:       " << blockOverlap    << " seconds"<< endl;
   }

   /////////////////////////////////////////////////////////////////////////////
   //                           create results directory                      //
   /////////////////////////////////////////////////////////////////////////////

   // check if (output directory exists
   if (!exist(outputDirectory,"dir")) {

      // report status
      wlog(debugLevel, 1, "creating output directory");

      // create output directory
      string shellcmd = "mkdir -p ";
      shellcmd += outputDirectory;
      int sysrc = system(shellcmd.c_str());
      if (sysrc == -1) {
	 cerr << "failed to fork command shell!" << endl;
      } else if (sysrc != 0) {
	 cerr << "failed to execute shell command: " << shellcmd << endl;
      }
   }

   ////////////////////////////////////////////////////////////////////////////
   //                              output file paths                         //
   ////////////////////////////////////////////////////////////////////////////

   // time string
   //string fileTimeString = startTimeString + "-" + segmentDurationString;
   ostringstream fileTimeString;
   fileTimeString <<  Now().getS() << "-" << pars.blockDuration;

   // generate output file path
   str_vect dummy_types;
   woutput outputFiles(pars.channelNames, fileTimeString.str(), 
		       outputDirectory, pars.triggerFormat, dummy_types);
   dummy_types.push_back("*");
   outputFiles.addtype(dummy_types, fileTimeString.str(), outputDirectory, 
		       "txt", "livetime");

   ////////////////////////////////////////////////////////////////////////////
   //                     Set up input data                                  //
   ////////////////////////////////////////////////////////////////////////////

   // report status
   wlog(debugLevel, 1, "Set up input stream");
   MultiDacc frameSource(frameCacheFile.c_str());
   // frameSource.setDebug(debugLevel);

   size_t nChannel = pars.channelNames.size();
   for (size_t i=0; i<nChannel; ++i) {
      frameSource.addChannel(pars.channelNames[i].c_str(), 0);
   }

   size_t nStates  = pars.stateNames.size();
   for (size_t i=0; i<nStates; ++i) {
      frameSource.addChannel(pars.stateNames[i].c_str(), 0);
   }

   size_t nInject  = pars.injectionNames.size();
   for (size_t i=0; i < nInject; ++i) {
      if (!pars.injectionNames[i].empty() && pars.injectionNames[i] != "NONE"){
	 frameSource.addChannel(pars.injectionNames[i].c_str(), 0);
      }
   }

   ////////////////////////////////////////////////////////////////////////////
   //                              begin block loop                          //
   ////////////////////////////////////////////////////////////////////////////
   wchunk block;

   // initialize error flag
   int errorFlag = 0;

   //  Get a boundary on an even number of nanoseconds (for time specification)
   //double boundary = gcd(fSample, 1000000000);

   // initialize coordinate variable
   dble_vect coordinate;
   size_t blockNumber=0;
   // begin loop over blocks
   while (!term) {
 
      // start time of block analysis
      EggTimer blockTimer;

      // report status
      if (debugLevel >= 1) {
	 cout << "analyzing block " << blockNumber+1 << " duration: "
	      << pars.blockDuration << " overlap: " << blockOverlap << endl; 
      }

      /////////////////////////////////////////////////////////////////////////
      //                            block analysis                           //
      /////////////////////////////////////////////////////////////////////////

      // try block and catch any errors
      str_vect channelNames;
      try {
	 block.process(pars.blockDuration, blockOverlap, tiling, pars, 
		       frameSource, outputFiles, debugLevel, channelNames);
      }

      catch (std::exception& e) {

	 if (debugLevel >= 1) {
	    cerr << "ERROR: " << e.what() << endl;
	 }

	 // note error, display, and reset
	 errorFlag += 1;

	 // continue to next block if (any errors found
	 continue;
      }

      /////////////////////////////////////////////////////////////////////////
      //                          report block time                          //
      /////////////////////////////////////////////////////////////////////////

      // report status
      double blockClockTime = blockTimer.elapsed();
      if (debugLevel >= 1) {
	 cout << "  block complete" << endl;
	 cout << "    elapsed time:          " << blockClockTime << " seconds"
	      << endl;

      }

      /////////////////////////////////////////////////////////////////////////
      //                          end block loop                             //
      /////////////////////////////////////////////////////////////////////////
      blockNumber +=1;
   }

   ////////////////////////////////////////////////////////////////////////////
   //                              write footer                              //
   ////////////////////////////////////////////////////////////////////////////

   // report status
   if (debugLevel >= 1) {
      cout << "segment complete" << endl;
      cout << "  elapsed time:            " << analysisTimer.elapsed() 
	   << " seconds" << endl;
   }

   ////////////////////////////////////////////////////////////////////////////
   //                                 return                                 //
   ////////////////////////////////////////////////////////////////////////////

   // exit with error status
   if (errorFlag) error("analysis had errors.");
}
